# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import os
import glob
import numpy as np
import torch
import torchvision as tv
import hdf5storage as hdf
import imageio
import matplotlib.pyplot as plt
from torch.utils.flop_counter import FlopCounterMode
from contextlib import nullcontext
import argparse
import time
import diffusers
import sys
from Utils import hsi


##########################################################


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--sensor',
                        default = 'aviris',
                        help='(string) sensor type (default = "aviris")')
    parser.add_argument('-sr', '--spatial_ratio',
                        type = int,
                        default = 8,
                        help = '(int) spatial subsampling ratio (default = 8)')
    parser.add_argument('-bb', '--band_block',
                        type = int,
                        default = 3,
                        help = '(int) number of bands to block together (default = 3)')
    parser.add_argument('-T', '--num_train_steps',
                        type = int,
                        default = 1000,
                        help = '(int) number of diffusion time steps (default = 1000)')
    parser.add_argument('-bmx', '--beta_max',
                        type = float,
                        default = 0.01,
                        help = '(float) maximum beta value (default = 0.01)')
    parser.add_argument('-bmn', '--beta_min',
                        type = float,
                        default = 0.0001,
                        help = '(float) minimum beta value (default = 0.0001)')
    parser.add_argument('-ns', '--num_inference_steps',
                        type = int,
                        default = 10,
                        help = '(int) number of inference time steps (default = 10)')
    parser.add_argument('-li', '--num_landweber_iterations',
                        type = int,
                        default = 2,
                        help = '(int) number of Landweber iterations (default = 5)')
    parser.add_argument('--ddim', action = 'store_true',
                        default = False,
                        help = 'use DDIM scheduler (default = DDPM)')
    parser.add_argument('--unconstrained', action = 'store_true',
                        default = False,
                        help = 'no enforcement of S or P constraints (implies Landweber iterations = 1)')
    parser.add_argument('--unconditioned', action = 'store_true',
                        default = False,
                        help = 'no conditioning')
    parser.add_argument('--transpose_adjoint', action = 'store_true',
                        default = False,
                        help = 'use transpose as spatial adjoint operator (default = bicubic interpolation)')
    parser.add_argument('--tau',
                        type = float,
                        default = None,
                        help = '(float) spectral regularizer constant (default = 3/2/sigma_1^2, where sigma_1 is the largest singular value of the S matrix)')
    parser.add_argument('--eta',
                        type = float,
                        default = 1.0,
                        help = '(float) spatial regularizer constant (default = 1.0')
    parser.add_argument('--flops', action = 'store_true',
                        default = False,
                        help = 'report number of FLOPS')
    parser.add_argument('testing_dir',
                        help = '(string) directory of testing images (MAT format)')
    parser.add_argument('model_dir',
                        help='(string) directory of trained model')
    parser.add_argument('results_dir',
                        help='(string) directory of output results')
    args = parser.parse_args()
    return args


##########################################################


def diffusion_update(sample, t, Lprime, H, model, scheduler,
                     num_bands, band_block, unconditioned):
    for band in range(0, num_bands, band_block):
        band_range = range(band, band + band_block)
        sample_block = sample[:, band_range, ...]
        if unconditioned:
            input = sample_block
        else:
            input = torch.cat((sample_block,
                               Lprime[:, band_range, ...],
                               H), dim = 1)
        with torch.inference_mode():
            noise_pred = model(input, t)["sample"]
        sample[:, band_range, ...] = scheduler.step(noise_pred, t,
                                                    sample_block).prev_sample
    return sample


def landweber_spectral_update(X, H, S, tau):
    residual = H - hsi.SpectralProjection(X, S)
    residual = hsi.SpectralProjection(residual, torch.t(S))
    return X + tau * residual

        
def landweber_spatial_update(X, L, P, eta, transpose_adjoint = False):
    spatial_ratio = P.shape[0]
    residual = L - hsi.SpatialProjection(X, P)
    if transpose_adjoint:
        residual = hsi.SpatialProjectionTranspose(residual, P)
    else:
        residual = hsi.BicubicInterpolation(residual, spatial_ratio)
    return X + eta * residual


def landweber(num_iterations, tau, eta, sample, t, Lprime, H, L, S, P,
              model, scheduler, num_bands, band_block,
              unconstrained = False,
              unconditioned = False,
              transpose_adjoint = False):
    if unconstrained:
        sample = diffusion_update(sample, t, Lprime, H,
                                  model, scheduler, num_bands, band_block,
                                  unconditioned)
    else:
        for i in range(num_iterations):
            sample = diffusion_update(sample, t, Lprime, H,
                                      model, scheduler, num_bands, band_block,
                                      unconditioned)
            sample = landweber_spectral_update(sample, H, S, tau)
            sample = landweber_spatial_update(sample, L, P, eta,
                                              transpose_adjoint)
    return sample


##########################################################


def CCDDReconstructImage(Lprime, H, L, S, P,
                         model_unet,
                         noise_scheduler,
                         tau,
                         eta,
                         band_block = 3,
                         num_landweber_iterations = 2,
                         unconstrained = False,
                         unconditioned = False,
                         transpose_adjoint = False,
                         device = None):
    
    if device is None:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    X_hat = torch.randn_like(Lprime)

    msi_num_bands, hsi_num_bands = S.shape

    for t in noise_scheduler.timesteps:
        X_hat = landweber(num_landweber_iterations, tau, eta,
                          X_hat, t, Lprime, H, L, S, P,
                          model_unet, noise_scheduler,
                          hsi_num_bands, band_block,
                          unconstrained,
                          unconditioned,
                          transpose_adjoint)

    return X_hat


##########################################################


def CCDDReconstructFile(filename,
                        sensor,
                        model_unet,
                        noise_scheduler,
                        results_dir = None,
                        output_error_image = False,
                        spatial_ratio = 8,
                        band_block = 3,
                        num_landweber_iterations = 2,
                        tau = None,
                        eta = 1.0,
                        unconstrained = False,
                        unconditioned = False,
                        transpose_adjoint = False,
                        flops = False,
                        device = None,
                        quiet = False):

    if device is None:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)
        
    S, remove_absorption_bands, get_pseudocolor_image = \
        hsi.ProcessSensor(sensor)

    if tau is None:
        sigma_1 = np.linalg.svd(S, full_matrices = False, compute_uv = False)[0]
        tau = 3/ 2 / sigma_1 ** 2
        
    S = torch.tensor(S, dtype = torch.float32, device = device)
    
    P = hsi.GaussianKernel(shape = (spatial_ratio, spatial_ratio))
    P = torch.tensor(P, dtype = torch.float32, device = device)

    if not os.path.isfile(filename):
        raise RuntimeError("Cannot find dataset %s" % filename)
    if not quiet: print('Processing image %s' % filename, flush = True)
        
    X = hdf.loadmat(filename)['X'].astype(np.float32)
    if remove_absorption_bands is not None:
        X = remove_absorption_bands(X)
    X = hsi.ToTensor(X, device = device)
    if X.shape[-3] != S.shape[1]:
        raise RuntimeError('Dataset %s (%d bands) does not match "%s" sensor (%d bands)' % (filename, X.shape[-3], sensor, S.shape[1]))
    num_rows = (X.shape[1] // spatial_ratio) * spatial_ratio
    num_cols = (X.shape[2] // spatial_ratio) * spatial_ratio
    if (num_rows != X.shape[1]) or (num_cols != X.shape[2]):
        if not quiet:
            print("Cropping %s to spatial size %d x %d" % (filename,
                                                           num_rows, num_cols),
                  flush = True)
        X = X[:, :num_rows, :num_cols]

    X, Xmax, Xmin = hsi.Normalize(X)
    X = X.expand(1, -1, -1, -1)

    # MSI
    H = hsi.SpectralProjection(X, S)
    
    # HSI
    L = hsi.SpatialProjection(X, P)
    Lprime = hsi.BicubicInterpolation(L, spatial_ratio)

    if flops:
        flop_counter = FlopCounterMode(display = False)
        
    tic = time.time()
    with flop_counter if flops else nullcontext():
        X_hat = CCDDReconstructImage(
            Lprime, H, L, S, P,
            model_unet = model_unet,
            noise_scheduler = noise_scheduler,
            tau = tau,
            eta = eta,
            band_block = band_block,
            num_landweber_iterations = num_landweber_iterations,
            unconstrained = unconstrained,
            unconditioned = unconditioned,
            transpose_adjoint = transpose_adjoint,
            device = device)
    toc = time.time()
    if not quiet: print('  Reconstruction time = %.2f (sec)' % (toc - tic),
                        flush = True)
    if flops and not quiet:
        print('  TFLOPS = %.2f' % 
              (flop_counter.get_total_flops() / 1e12), flush = True)
    
    X = torch.squeeze(X, 0)
    X_hat = torch.squeeze(X_hat, 0)

    # Renormalize to [0, 1] for quality measures, for consistency with LRTA
    X = X / 2 + 0.5
    X_hat = X_hat / 2 + 0.5

    psnr = hsi.PSNR(X, X_hat, max_val = 1.0)
    rmse = hsi.RMSE(X, X_hat)
    sam = hsi.SAM(X, X_hat)
    if not quiet:
        print('    PSNR = %.2f dB' % psnr, flush = True)
        print('    RMSE = %e' % rmse, flush = True)
        print('     SAM = %.4f deg' % sam, flush = True)

    if results_dir is not None:
        outfile = os.path.splitext(os.path.basename(filename))[0]
        outimagefile = os.path.join(results_dir,
                                    outfile + ".reconstructed.mat")
        if not quiet:
            print('  Saving reconstructed image as %s...' % outimagefile,
                  flush = True)
        
        if os.path.isfile(outimagefile):
            os.remove(outimagefile)
        hdf.savemat(outimagefile, {'X_hat': hsi.ToNumpy(X_hat)})
    
        if output_error_image:
            errplotfile = os.path.join(results_dir,
                                       outfile + ".error.png")
            if not quiet:
                print('  Saving error image as %s...' % errplotfile,
                      flush = True)
            hsi.SaveErrorPlot(X, X_hat, errplotfile)
    
        if get_pseudocolor_image is not None:
            outimagefile = os.path.join(results_dir,
                                        outfile + ".reconstructed.png")
            X_hat = hsi.Denormalize(X_hat, Xmax, Xmin, type = "minmax")
            if not quiet:
                print('  Saving pseudocolor image as %s...' % outimagefile,
                      flush = True)
            outimage = get_pseudocolor_image(hsi.ToNumpy(X_hat))
            imageio.imwrite(outimagefile, outimage)
            
    return psnr, rmse, sam


##########################################################


def CCDDReconstruct(testing_dir,
                    model_dir,
                    results_dir,
                    sensor,
                    spatial_ratio = 8,
                    band_block = 3,
                    num_train_steps = 1000,
                    beta_max = 0.01,
                    beta_min = 0.0001,
                    num_inference_steps = 10,
                    num_landweber_iterations = 2,
                    tau = None,
                    eta = 1.0,
                    ddim = False,
                    unconstrained = False,
                    unconditioned = False,
                    transpose_adjoint= False,
                    flops = False,
                    device = None,
                    quiet = False):

    if not os.path.isdir(testing_dir):
        raise RuntimeError("Cannot find testing direcoty %s" % testing_dir)
    if not os.path.isdir(model_dir):
        raise RuntimeError("Cannot find model direcoty %s" % model_dir)
    os.makedirs(results_dir, exist_ok = True)
    
    if unconditioned and unconstrained:
        raise RuntimeError("Cannot have unconditioned and unconstrained reconstruction simultaneously")
        
    if device is None:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)
        
    if unconstrained:
        num_landweber_iterations = 1

    if not quiet:
        print('CCDD testing (%s) - parameters:' % device, flush = True)
        if ddim:
            print('  Scheduler: DDIM', flush = True)
        else:
            print('  Scheduler: DDPM', flush = True)
        print('  Num inference steps: %d' % num_inference_steps, flush = True)
        if unconditioned:
            print('  Conditioning: unconditioned', flush = True)
        else:
            print('  Conditioning: conditioned', flush = True)
        if unconstrained:
            print('  Constraints: unconstrained', flush = True)
            print('  Num Landweber iterations: not using S or P constraints', flush = True)
        else:
            print('  Constraints: constrained', flush = True)
            print('  Num Landweber iterations: %d' % num_landweber_iterations, flush = True)
        if transpose_adjoint:
            print('  Spatial adjoint: transpose', flush = True)
        else:
            print('  Adjoint: bicubic interpolation', flush = True)
        if tau is None:
            print('  tau: tau_star', flush = True)
        else:
            print('  tau: %f' % tau, flush = True)
        print('  eta: %f\n' % eta, flush = True)
        
    # X: HSI treated as HS2I
    testing_files = sorted(glob.glob(os.path.join(testing_dir, '*.mat')))
    
    if not quiet: print('Loading model from %s...' % model_dir, flush = True)
    model_unet = diffusers.UNet2DModel.from_pretrained(model_dir,
                                                       use_safetensors = False,
                                                       device_map = device)
    num_parameters = sum(p.numel() for p in model_unet.parameters())
    if not quiet: print('  Number of UNet parameters: %.2f (million)\n' % \
                        (num_parameters / 1e6),
          flush = True)
    
    if ddim:
        noise_scheduler = diffusers.DDIMScheduler(num_train_timesteps =
                                                  num_train_steps,
                                                  beta_start = beta_min,
                                                  beta_end = beta_max)
    else:
        noise_scheduler = diffusers.DDPMScheduler(num_train_timesteps =
                                              num_train_steps,
                                              beta_start = beta_min,
                                              beta_end = beta_max)
    
    noise_scheduler.set_timesteps(num_inference_steps = num_inference_steps)

    psnrs = []
    rmses = []
    sams = []
    
    for filename in testing_files:
        psnr, rmse, sam = \
            CCDDReconstructFile(
                filename = filename,
                sensor = sensor,
                model_unet = model_unet,
                noise_scheduler = noise_scheduler,
                results_dir = results_dir,
                output_error_image = True,
                spatial_ratio = spatial_ratio,
                band_block = band_block,
                num_landweber_iterations = num_landweber_iterations,
                tau = tau,
                eta = eta,
                unconstrained = unconstrained,
                unconditioned = unconditioned,
                transpose_adjoint = transpose_adjoint,
                flops = flops,
                device = device,
                quiet = quiet)
        psnrs.append(psnr)
        rmses.append(rmse)
        sams.append(sam)

    return psnrs, rmses, sams



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

# For data from the Headwall Hyperspec-VNIR-C imaging sensor as described by:
#   N. Yokoya and A. Iwasaki, "Airborne hyperspectral data over Chikusei,"
#   Space Appl. Lab., Univ. Tokyo, Japan, Tech. Rep. SAL-2016-05-27, May 2016.

import numpy as np
from PIL import Image
from PIL import ImageEnhance

# VNIR 224 bands total
num_bands = 128
# VNIR radiometric resolution = 14 bits
bit_depth = 14


# VNIR wavelengths: 363 to 1018 nm
wavelength = np.linspace(363, 1018, num_bands)


# VNIR water-absorption bands: none, but get rid of last two bands for
# divisibilty by 3 needed by CCDD
def RemoveAbsorptionBands(X):
    if X.shape[-1] == num_bands:
        return X[..., :-2]
    else:
        return X


# Four multispectral bands of Landsat TM (skips Band 5, 6, 7)
# Bands                                       Wavelength (micrometers)
# ------------------------------------------------------------------
# Band 1 - Blue                                0.45 -  0.52
# Band 2 - Green                               0.52 -  0.60
# Band 3 - Red                                 0.63 -  0.69
# Band 4 - NIR                                 0.76 -  0.90
# Band 5 - Shortwave Infrared (SWIR) 1         1.55 -  1.75 
# Band 6 - Thermal                            10.40 - 12.50
# Band 7 - SWIR 2                              2.08 -  2.35
def GetSpectralProjection():
    band1 = ((450 <= wavelength) & (wavelength <= 520)).reshape(1, num_bands)
    band2 = ((520 <= wavelength) & (wavelength <= 600)).reshape(1, num_bands)
    band3 = ((630 <= wavelength) & (wavelength <= 690)).reshape(1, num_bands)
    band4 = ((760 <= wavelength) & (wavelength <= 900)).reshape(1, num_bands)
    
    S = np.concatenate((band1, band2, band3, band4), axis=0).astype(np.float32)
    return (S / np.sum(S, axis=1).reshape(4, 1))[:, :-2].astype(np.float32)


def PseudocolorBands():
    return (52, 33, 20)


def GetPseudocolorImage(X):
    if X.shape[-1] == num_bands:
        X = RemoveAbsorptionBands(X)
    X = X[..., PseudocolorBands()]
    X = X / (2. ** bit_depth) * 255.
    X = np.clip(X, 0, 255)
    X = X.astype(np.uint8)
    X = Image.fromarray(X)
    return ImageEnhance.Brightness(X).enhance(15)



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import glob
import numpy as np
import torch
import torchvision as tv
import hdf5storage as hdf
import time
import datasets
import diffusers
import bitsandbytes as bnb
from Utils import hsi, aviris, rosis, vnir


##########################################################


class hsi_dataset(torch.utils.data.Dataset):
    def __init__(self, X, H, Y):
        self.X = X
        self.H = H
        self.Y = Y
    def __len__(self):
        return self.X.shape[0]
    def __getitem__(self, i):
        return {"X": self.X[i],
                "H": self.H[i],
                "Y": self.Y[i]}

    
def get_patch(X, S, patch_size, band_block):
    num_bands = X.shape[0]
    X = tv.transforms.RandomCrop(patch_size)(X)
    H = hsi.SpectralProjection(X, S)
    start_band = np.random.randint(num_bands // band_block) * band_block
    return X[start_band:(start_band + band_block), :, :], H


##########################################################


def CCDDTrainLoop(dataloader,
                  model_unet,
                  optimizer,
                  noise_scheduler,
                  start_epoch = 0,
                  end_epoch = 300,
                  save_epochs = 10,
                  model_dir = None,
                  optimizer_file = None,
                  history_file = None,
                  unconditioned = False,
                  device = None,
                  quiet = False):

    losses = []
    num_train_timesteps = noise_scheduler.config.num_train_timesteps

    for epoch in range(start_epoch, end_epoch):
        epoch_tic = time.time()
        num_batches = len(dataloader)
        for batch_cnt, batch in enumerate(dataloader, start = 1):
            X = batch["X"].to(device)
            H = batch["H"].to(device)
            Y = batch["Y"].to(device)
            noise = torch.randn(X.shape, device = device)
            timesteps = torch.randint(0,
                                      num_train_timesteps,
                                      (X.shape[0],), device = device).long()
            X_hat = noise_scheduler.add_noise(X, noise, timesteps)
            if unconditioned:
                input = X_hat
            else:
                input = torch.cat((X_hat, Y, H), dim = 1)
            noise_hat = model_unet(input, timesteps, return_dict = False)[0]
            loss = torch.nn.functional.mse_loss(noise_hat, noise)
            losses.append(loss.item())
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
    
        epoch_toc = time.time()
        avg_loss = sum(losses[-num_batches:]) / num_batches
        epoch_time = epoch_toc - epoch_tic
        if not quiet:
            print("Epoch: %d -- loss: %f -- time %.4f (sec)" % (epoch,
                                                                avg_loss,
                                                                epoch_time),
              flush = True)
        if history_file is not None:
            with open(history_file, "a") as outfile:
                outfile.write("%f\n" % avg_loss)
        if (epoch + 1) % save_epochs == 0 and model_dir is not None:
            if not quiet:
                print("Saving model to %s..." % model_dir, flush = True)
            model_unet.save_pretrained(model_dir, safe_serialization = False)
            if optimizer_file is not None:
                torch.save(optimizer.state_dict(), optimizer_file)

    return model_unet, optimizer, losses


##########################################################


def CCDDTrain(training_dir,
              model_dir,
              sensor,
              spatial_ratio = 8,
              band_block = 3,
              num_train_steps = 1000,
              beta_max = 0.01,
              beta_min = 0.0001,
              learning_rate = 0.0001,
              batch_size = 64,
              patch_size = 32,
              num_patches = 8192,
              unet_channels = 128,
              unet_layers = 3,
              start_epoch = 0,
              end_epoch = 300,
              save_epochs = 10,
              history_file = None,
              unconditioned = False,
              device = None,
              quiet = False):

    if device is None:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    if not os.path.isdir(training_dir):
        raise RuntimeError("Cannot find training direcotry %s" % training_dir)
    os.makedirs(model_dir, exist_ok = True)
    
    
    S, remove_absorption_bands, _ = hsi.ProcessSensor(sensor)

    S = torch.tensor(S, dtype = torch.float32, device = device)
    
    P = hsi.GaussianKernel(shape = (spatial_ratio, spatial_ratio))
    P = torch.tensor(P, dtype = torch.float32, device = device)
    
    resume_training = start_epoch != 0
    
    if not quiet:
        print('CCDD training (%s) - parameters:' % device, flush = True)
        print('  Num training time steps (T): %d' % num_train_steps,
              flush = True)
        print('  Beta max: %f' % beta_max, flush = True)
        print('  Beta min: %f' % beta_min, flush = True)
        print('  Learning rate: %f' % learning_rate, flush = True)
        print('  Batch size: %d' % batch_size, flush = True)
        print('  Patch size: %d' % patch_size, flush = True)
        print('  Num patches: %d' % num_patches, flush = True)
        print('  Band block: %d' % band_block, flush = True)
        print('  Initial epoch: %d' % start_epoch, flush = True)
        print('  End epoch: %d' % end_epoch, flush = True)
        print('  Num UNet input channels: %d' % unet_channels, flush = True)
        print('  Num UNet layers: %d' % unet_layers, flush = True)
        if unconditioned:
            print('  Conditioning: unconditioned', flush = True)
        print('', flush = True)
        print('Data parameters:', flush = True)
        print('  Sensor: %s' % sensor, flush = True)
        print('  Spatial ratio: %d' % spatial_ratio, flush = True)
        print('', flush = True)

    # X: HSI treated as HS2I
    if not quiet: print("Loading data...", flush = True)
    training_files = sorted(glob.glob(os.path.join(training_dir, '*.mat')))
    
    X = []
    H = []
    Y = []
    for training_file in training_files:
        X_current = hdf.loadmat(training_file)['X'].astype(np.float32)
        if remove_absorption_bands is not None:
            X_current = remove_absorption_bands(X_current)
        X_current = hsi.ToTensor(X_current, device = device)
        X_current = hsi.Normalize(X_current)[0]
        if X_current.shape[-3] != S.shape[1]:
            raise RuntimeError('Dataset %s (%d bands) does not match "%s" sensor (%d bands)' % (training_file, X_current.shape[-3], sensor, S.shape[1]))
        for patch in range(num_patches):
            current_patch, H_current = get_patch(X_current, S,
                                                 patch_size, band_block)
            L_current = hsi.SpatialProjection(current_patch, P)
            Y_current  = hsi.BicubicInterpolation(L_current, spatial_ratio)
            X.append(current_patch)
            H.append(H_current)
            Y.append(Y_current)
            
    dataset = hsi_dataset(torch.stack(X),
                          torch.stack(H),
                          torch.stack(Y))
    
    msi_num_bands = S.shape[0]
    
    dataloader = torch.utils.data.DataLoader(dataset,
                                             batch_size = batch_size,
                                             shuffle = True)
    
    noise_scheduler = diffusers.DDPMScheduler(num_train_timesteps =
                                              num_train_steps,
                                              beta_start = beta_min,
                                              beta_end = beta_max)
    
    if not resume_training:
        channels = unet_channels
        block_out_channels = (channels,) + tuple((channels := 2 * channels
                                                  for layer in
                                                  range(unet_layers - 1)))
        down_block_types = ("DownBlock2D",) * unet_layers
        up_block_types = ("UpBlock2D",) * unet_layers
        if unconditioned:
            in_channels = band_block
        else:
            in_channels = 2 * band_block + msi_num_bands
        model_unet = diffusers.UNet2DModel(
            in_channels = in_channels,
            out_channels = band_block,
            block_out_channels = block_out_channels,
            down_block_types = down_block_types,
            up_block_types = up_block_types)
        model_unet = model_unet.to(device)
    else:
        model_unet = diffusers.UNet2DModel.from_pretrained(
            model_dir,
            use_safetensors = False,
            device_map = device)

    model_unet.enable_gradient_checkpointing()
    
    optimizer = bnb.optim.Adam8bit(model_unet.parameters(), lr = learning_rate)
    optimizer_file = os.path.join(model_dir, "optimizer.pt")
    if resume_training:
        optimizer.load_state_dict(torch.load(optimizer_file))
    
    if history_file is not None:
        if os.path.exists(history_file) and not resume_training:
            os.remove(history_file)

    if not quiet: print("Training model...", flush = True)
    training_tic = time.time()
    model_unet, optimizer, losses = \
        CCDDTrainLoop(dataloader = dataloader,
                      model_unet = model_unet,
                      optimizer = optimizer,
                      noise_scheduler = noise_scheduler,
                      start_epoch = start_epoch,
                      end_epoch = end_epoch,
                      save_epochs = save_epochs,
                      model_dir = model_dir,
                      optimizer_file = optimizer_file,
                      history_file = history_file,
                      unconditioned = unconditioned,
                      device = device,
                      quiet = quiet)
    training_toc = time.time()
    training_time = training_toc - training_tic
    if not quiet:
        print('Training time = %.2f (sec)' % training_time, flush = True)
    
    if not quiet:
        print("Saving model to %s..." % model_dir, flush = True)
    model_unet.save_pretrained(model_dir, safe_serialization = False)
    torch.save(optimizer.state_dict(), optimizer_file)

    return model_unet, optimizer, losses



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import tempfile
import tarfile
import shutil
import subprocess
import numpy as np
import hdf5storage as hdf
import imageio
import xlrd
from Utils import aviris

def MakeData():
    num_rows_extracted = 256
    num_cols_extracted = 256
    
    dirname = os.path.dirname(os.path.realpath(__file__))

    xls_file = os.path.join(dirname, "losangeles.xls")
    training_dir = os.path.join(dirname, "Training")
    testing_dir = os.path.join(dirname, "Testing")
    os.makedirs(training_dir, exist_ok = True)
    os.makedirs(testing_dir, exist_ok = True)
    
    workbook = xlrd.open_workbook(xls_file)
    worksheet = workbook.sheet_by_index(0)

    with tempfile.TemporaryDirectory() as download_dir:
        for row in range(1, worksheet.nrows):
            datasetname = worksheet.cell_value(row, 1)
            fileurl = worksheet.cell_value(row, 38)
            num_rows = int(worksheet.cell_value(row, 22))
            num_cols = int(worksheet.cell_value(row, 23))
            num_bands = 224
        
            row_origin = int(worksheet.cell_value(row, 51)) - 1
            col_origin = int(worksheet.cell_value(row, 52)) - 1
        
            training = bool(worksheet.cell_value(row, 53))
        
            filename = worksheet.cell_value(row, 54)
            
            dir = training_dir if training else testing_dir

            ext = '.training.mat' if training else '.testing.mat'
            matfile = os.path.join(dir, filename + ext)

            if os.path.exists(matfile):
                print("  Nothing to do - %s exists already" % matfile)
                continue
            
            radiance_version = worksheet.cell_value(row, 7)
        
            basename = os.path.basename(fileurl)
            downloadfilename = os.path.join(download_dir,
                                            os.path.basename(fileurl))
            dirname1  = os.path.splitext(os.path.splitext(basename)[0])[0]
            dirname2  = os.path.splitext(os.path.splitext(basename)[0])[0] + \
                'rdn_' + radiance_version
            
            print('Dataset %s, rows = %d, cols = %d' % (datasetname,
                                                        num_rows, num_cols),
                  flush = True)
            
            print('  Downloading %s as %s...' % (datasetname,
                                                 downloadfilename),
                  flush = True)
            if subprocess.run(['wget', '-q', '-P', download_dir, fileurl]):
                print("Error downloading %s" % fileurl)
                continue

            with tarfile.open(downloadfilename, mode = 'r:gz') as tf:
                # Find directory, using only first scene
                names = tf.getnames()
                filename1 = "%s/%s_sc01_ort_img" % (dirname1, dirname1)
                filename2 = "%s/%s_sc01_ort_img" % (dirname2, dirname2)
                if filename1 in names:
                    print("  Found %s" % filename1)
                    scenefile = filename1
                elif filename2 in names:
                    print("  Found %s" % filename2)
                    scenefile = filename2
                else:
                    print('  Cannot find directory and/or ort_img file')
                    continue
                print('  Processing %s as %s : %dx%d at (%d, %d)...' % 
                      (scenefile, filename,
                       num_rows_extracted, num_cols_extracted,
                       row_origin, col_origin),
                      flush = True)
                tf.extract(scenefile, path = download_dir)

            scenefile = os.path.join(download_dir, scenefile)
            with open(scenefile, 'r') as infile:
                X = np.fromfile(infile, dtype='>i2').reshape(num_rows,
                                                             num_cols,
                                                             num_bands,
                                                             order='C')
                
            X = X[row_origin:(row_origin + num_rows_extracted),
                  col_origin:(col_origin + num_cols_extracted), :]
        
            print('  Writing %s...' % matfile, flush = True)
            hdf.savemat(matfile, {'X': X})
        
            ext = '.training.png' if training else '.testing.png'
            imagefile = os.path.join(dir, filename + ext)
            image = aviris.GetPseudocolorImage(X)
            imageio.imwrite(imagefile, image)

            os.remove(downloadfilename)
            os.remove(scenefile)



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import numpy as np
from PIL import Image
from PIL import ImageEnhance

# AVIRIS 224 bands total
num_bands = 224
# AVIRIS radiometric resolution = 14 bits
bit_depth = 14


# AVIRIS wavelengths: 400 to 2500 nm
wavelength = np.linspace(400, 2500, num_bands)


# AVIRIS water-absorption bands 108-112, 154-167, and 224
water_absorption_bands = np.array([band for band in range(108, 113)] +
                                  [band for band in range(154, 168)] +
                                  [224]) - 1


def AbsorptionBoolean():
    I = np.full((num_bands), True)
    I[water_absorption_bands] = False
    return I


def RemoveAbsorptionBands(X):
    if X.shape[-1] == num_bands:
        return X[..., AbsorptionBoolean()]
    else:
        return X


# Six multispectral bands of Landsat TM (skips Band 6)
# Bands                                       Wavelength (micrometers)
# ------------------------------------------------------------------
# Band 1 - Blue                                0.45 -  0.52
# Band 2 - Green                               0.52 -  0.60
# Band 3 - Red                                 0.63 -  0.69
# Band 4 - NIR                                 0.76 -  0.90
# Band 5 - Shortwave Infrared (SWIR) 1         1.55 -  1.75 
# Band 6 - Thermal                            10.40 - 12.50
# Band 7 - SWIR 2                              2.08 -  2.35
def GetSpectralProjection():
    band1 = ((450 <= wavelength) & (wavelength <= 520)).reshape(1, num_bands)
    band2 = ((520 <= wavelength) & (wavelength <= 600)).reshape(1, num_bands)
    band3 = ((630 <= wavelength) & (wavelength <= 690)).reshape(1, num_bands)
    band4 = ((760 <= wavelength) & (wavelength <= 900)).reshape(1, num_bands)
    band5 = ((1550 <= wavelength) & (wavelength <= 1750)).reshape(1, num_bands)
    band7 = ((2080 <= wavelength) & (wavelength <= 2350)).reshape(1, num_bands)
    
    S = np.concatenate((band1, band2, band3, band4, band5, band7), axis=0).astype(np.float32)
    return (S / np.sum(S, axis=1).reshape(6, 1))[:, AbsorptionBoolean()].astype(np.float32)


def PseudocolorBands():
    return (28, 16, 8)


def GetPseudocolorImage(X):
    if X.shape[-1] == num_bands:
        X = RemoveAbsorptionBands(X)
    X = X[..., PseudocolorBands()]
    X = X / (2. ** bit_depth) * 255.
    X = np.clip(X, 0, 255)
    X = X.astype(np.uint8)
    X = Image.fromarray(X)
    return ImageEnhance.Brightness(X).enhance(2.75)



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import numpy as np
import numpy.linalg as npla
import numpy.matlib as ml
import scipy
import matplotlib.pyplot as plt
import time
from PIL import Image
import torch
import torch.nn.functional as F
import imageio
import tiler
# from torcheval.metrics import StructuralSimilarity as SSIM
# import torcheval.metrics.image.ssim as tm
import skimage.metrics as skim
from Utils import aviris, aviris220, rosis, vnir, hydice


###########################################################################
#
# HSI images in ndarray form have format (channels-last):
#    (batch_size, num_rows, num_cols, num_bands)
#
# HSI images in torch.tensor form have format (channels-first):
#    (batch_size, num_bands, num_rows, num_cols)
#
# hsi.ToTensor() and hsi.ToNumpy() automatically convert between
# channels-first and channels-last formats
#
###########################################################################


def ToTensor(X, dtype = None, device = None):
    if torch.is_tensor(X):
        return X
    X = torch.tensor(X, dtype = dtype, device = device)
    if X.ndim == 4:
        X = torch.permute(X, (0, 3, 1, 2))
    else:
        X = torch.permute(X, (2, 0, 1))
    return X


def ToNumpy(X):
    if not torch.is_tensor(X):
        return X
    if X.ndim == 4:
        X = torch.permute(X, (0, 2, 3, 1))
    else:
        X = torch.permute(X, (1, 2, 0))
    return np.array(X.cpu())


def Normalize(X, type = None):
    if torch.is_tensor(X):
        Xmax = torch.max(X)
        Xmin = torch.min(X)
    else:
        Xmax = np.max(X)
        Xmin = np.min(X)
    X = (X - Xmin) / (Xmax - Xmin)
    if type == "minmax":
        return X, Xmax, Xmin
    else:
        return (X - 0.5) * 2, Xmax, Xmin


def Denormalize(X, Xmax, Xmin, type = None):
    if type != "minmax":
        X = X / 2 + 0.5
    X = X * (Xmax - Xmin) + Xmin
    return X

    
def GaussianKernel(shape = (8, 8), sigma = 2.0):
    m, n = [(ss - 1.)/2. for ss in shape]
    y,x = np.ogrid[-m:m+1,-n:n+1]
    h = np.exp( -(x*x + y*y) / (2. * sigma * sigma) )
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h


# Use 'C' or 'torch' for row-major order, 'F' or 'matlab' for column-major order
def GetSpatialProjectionAsMatrix(spatial_ratio, num_rows, num_cols,
                                 psf = None,
                                 sparse = False,
                                 tensor = False,
                                 order = 'C'):
    if psf is None:
        psf = GaussianKernel(shape = (spatial_ratio, spatial_ratio))
    m, n = psf.shape
    M = num_rows // spatial_ratio
    N = num_cols // spatial_ratio
    num = m * n
    NUM = M * N

    if order == 'F' or order == 'matlab':
        I = np.kron(np.arange(1, NUM + 1).reshape(NUM, 1),
                    np.ones((num, 1))).astype(int).flatten() - 1
        Temp1 = np.kron(num_rows * np.arange(0, n),
                        np.ones((1, m))); 
        Temp2 = ml.repmat(np.arange(1, m + 1), 1, n)
        Temp3 = Temp1 + Temp2
        
        X, Y = np.meshgrid(np.arange(0, M), np.arange(0, N))
        Temp4 = (X * m + Y * num_rows * n).reshape(M * N, 1)
        Temp5 = ml.repmat(Temp3, 1, NUM)
        Temp6 = np.kron(Temp4.T, np.ones((1, num)))
        J = (Temp5 + Temp6).astype(int).flatten() - 1
    elif order == 'C' or order == 'torch':
        I = np.kron(np.arange(1, NUM + 1).reshape(NUM, 1),
                    np.ones((num, 1))).astype(int).flatten() - 1
        Temp1 = np.kron(num_cols * np.arange(0, m),
                        np.ones((1, n))); 
        Temp2 = ml.repmat(np.arange(1, n + 1), 1, m)
        Temp3 = Temp1 + Temp2
        
        Y, X = np.meshgrid(np.arange(0, N), np.arange(0, M))
        Temp4 = (Y * n + X * num_cols * m).reshape(NUM, 1)
        Temp5 = ml.repmat(Temp3, 1, NUM)
        Temp6 = np.kron(Temp4.T, np.ones((1, num)))
        J = (Temp5 + Temp6).astype(int).flatten() - 1
    else:
        print("Invalid order (%s)" % order)
        return None
        
    P = ml.repmat(psf.reshape(num, 1), NUM, 1).flatten()

    P = scipy.sparse.coo_matrix((P, (I, J)), dtype=np.float32).T
    
    if not sparse:
        if tensor:
            P = torch.tensor(P)
        else:
            P = P.toarray()
    else:
        if tensor:
            values = torch.FloatTensor(P.data)
            indices = torch.LongTensor(np.vstack((P.row, P.col)))
            P = torch.sparse.FloatTensor(indices, values,
                                         torch.Size(P.shape))
    return P


# def GetSpatialProjectionAsMatrix(spatial_ratio, num_rows, num_cols,
#                                  psf = None,
#                                  sparse = False,
#                                  tensor = False):
#     if psf is None:
#         psf = GaussianKernel(shape = (spatial_ratio, spatial_ratio))
#     m, n = psf.shape
#     M = num_rows // spatial_ratio
#     N = num_cols // spatial_ratio
#     num = m * n
#     NUM = M * N
#     I = np.kron(np.arange(1, NUM + 1).reshape(NUM, 1),
#                 np.ones((num, 1))).astype(int).flatten() - 1
#     Temp1 = np.kron(num_cols * np.arange(0, m),
#                     np.ones((1, n))); 
#     Temp2 = ml.repmat(np.arange(1, n + 1), 1, m)
#     Temp3 = Temp1 + Temp2
# 
#     Y, X = np.meshgrid(np.arange(0, N), np.arange(0, M))
#     Temp4 = (Y * n + X * num_cols * m).reshape(NUM, 1)
#     Temp5 = ml.repmat(Temp3, 1, NUM)
#     Temp6 = np.kron(Temp4.T, np.ones((1, num)))
#     J = (Temp5 + Temp6).astype(int).flatten() - 1
#     
#     P = ml.repmat(psf.reshape(num, 1), NUM, 1).flatten()
# 
#     P = scipy.sparse.coo_matrix((P, (I, J)), dtype=np.float32).T
#     
#     if not sparse:
#         if tensor:
#             P = torch.tensor(P)
#         else:
#             P = P.toarray()
#     else:
#         if tensor:
#             values = torch.FloatTensor(P.data)
#             indices = torch.LongTensor(np.vstack((P.row, P.col)))
#             P = torch.sparse.FloatTensor(indices, values,
#                                          torch.Size(P.shape))
#     return P


def SpectralProjectionUnfolded(X3, S):
    return S @ X3


def SpectralProjection(X, S):
    if torch.is_tensor(X):
        num_rows = X.shape[-2]
        num_cols = X.shape[-1]
    else:
        num_rows = X.shape[-3]
        num_cols = X.shape[-2]
    X3 = Unfold3(X)
    return Fold3(SpectralProjectionUnfolded(X3, S), num_rows, num_cols)

    
def SpatialProjectionUnfolded(X3, P):
    return X3 @ P


def SpatialProjection(X, P, strides = None, padding = 0):
    return_numpy = False
    if not torch.is_tensor(X):
        X = ToTensor(X)
        return_numpy = True
    if not torch.is_tensor(P):
        P = torch.tensor(P)
        
    return_squeeze = False
    if X.dim() == 3:
        X = torch.unsqueeze(X, 0)
        return_squeeze = True

    if strides == None:
        strides = (P.shape[0], P.shape[1])
    num_bands = X.shape[1]
    P = P.expand([num_bands, 1, -1, -1])

    X_tilde = F.conv2d(X, P, stride = strides, groups = num_bands,
                       padding = padding)
    if return_squeeze:
        X_tilde = torch.squeeze(X_tilde, 0)

    if return_numpy:
        return ToNumpy(X_tilde)
    else:
        return X_tilde
            
#     else:
#         return_squeeze = False
#         if (len(X.shape)) == 3:
#             X = X[np.newaxis, :, :, :]
#             return_squeeze = True
# 
#         num_batches = X.shape[0]
#         num_bands = X.shape[3]
#         P = np.tile(P[:, :, np.newaxis, np.newaxis], (1, 1, hsi_num_bands, 1))
#         num_out_channels = P.shape[3]
# 
#         X_tilde = np.zeros((num_batches, X.shape[1], X.shape[2],
#                             num_bands * num_out_channels), dtype = np.float32)
#         for batch in range(num_batches):
#             for band in range(num_bands):
#                 for channel in range(num_out_channels):
#                     out_band = band * num_out_channels + channel
#                     X_tilde[batch, :, :, out_band] = \
#                         scipy.ndimage.correlate(X[batch, :, :, band].squeeze(),
#                                                 P[:, :, band, channel].squeeze(),
#                                                 mode = 'constant')
# 
#         X_tilde = X_tilde[:,
#                           (P.shape[0] // 2)::P.shape[0],
#                           (P.shape[1] // 2)::P.shape[1], :]
#                                             
#         if return_squeeze:
#             X_tilde = np.squeeze(X_tilde, 0)


    
def SpatialProjectionTranspose(X_tilde, P, strides = None, padding = 0):
    return_numpy = False
    if not torch.is_tensor(X_tilde):
        X_tilde = ToTensor(X_tilde)
        return_numpy = True
    if not torch.is_tensor(P):
        P = torch.tensor(P)

    return_squeeze = False
    if X_tilde.dim() == 3:
        X_tilde = torch.unsqueeze(X_tilde, 0)
        return_squeeze = True
        
    if strides == None:
        strides = (P.shape[0], P.shape[1])
    num_bands = X_tilde.shape[1]
    P = P.expand([num_bands, 1, -1, -1])
    
    X = F.conv_transpose2d(X_tilde, P, stride = strides, groups = num_bands,
                           padding = 0)
    if return_squeeze:
        X = torch.squeeze(X, 0)

    if return_numpy:
        return ToNumpy(X)
    else:
        return X

    
# def SpatialProjectionTranspose(X_tilde, P, num_rows, num_cols):
#     return_numpy = False
#     if not torch.is_tensor(X_tilde):
#         X_tilde = tf.convert_to_tensor(X_tilde)
#         return_numpy = True
#         
#     return_squeeze = False
#     if (len(X_tilde.shape)) == 3:
#         X_tilde = tf.expand_dims(X_tilde, 0)
#         return_squeeze = True
# 
#     strides = [1, P.shape[0], P.shape[1], 1]
#     shape = [tf.shape(X_tilde)[0], num_rows, num_cols, tf.shape(X_tilde)[3]]
#     shape = tf.stack(shape)
#     X = tf.nn.depthwise_conv2d_backprop_input(shape,
#                                               P, X_tilde,
#                                               strides = strides,
#                                               padding = 'SAME')
# 
#     if return_squeeze:
#         X = tf.squeeze(X, 0)
#         
#     if return_numpy:
#         return X.numpy()
#     else:
#         return X


def Unfold1(X):
    if torch.is_tensor(X):
        if X.ndim == 4:
            batch_size, num_bands, num_rows, num_cols = X.shape
            X = torch.permute(X, (0, 2, 3, 1))
#             X = torch.permute(X, (0, 2, 1, 3))
            return X.reshape(-1, num_rows, num_cols * num_bands)
        else:
            num_bands, num_rows, num_cols = X.shape
            X = torch.permute(X, (1, 2, 0))
#             X = torch.permute(X, (1, 0, 2))
            return X.reshape(num_rows, num_cols * num_bands)
    else:
        if X.ndim == 4:
            batch_size, num_rows, num_cols, num_bands = X.shape
            return X.reshape(-1, num_rows, num_cols * num_bands)
        else:
            num_rows, num_cols, num_bands = X.shape
            return X.reshape(num_rows, num_cols * num_bands)
        
    
def Fold1(X1, num_cols, num_bands):
    if X1.ndim == 3:
        num_rows = X1.shape[1]
    else:
        num_rows = X1.shape[0]
        
    if torch.is_tensor(X1):
        if X1.ndim == 3:
            X = X1.reshape(-1, num_rows, num_cols, num_bands)
            X = torch.permute(X, (0, 3, 1, 2))
            return X
#             X = X1.reshape(-1, num_rows, num_bands, num_cols)
#             return torch.permute(X, (0, 2, 1, 3))
        else:
            X = X1.reshape(num_rows, num_cols, num_bands)
            X = torch.permute(X, (2, 0, 1))
            return X
#             X = X1.reshape(num_rows, num_bands, num_cols)
#             return torch.permute(X, (1, 0, 2))
    else:
        if X1.ndim == 3:
            X = X1.reshape(-1, num_rows, num_cols, num_bands)
            return X
        else:
            X = X1.reshape(num_rows, num_cols, num_bands)
            return X

        
def Unfold2(X):
    if torch.is_tensor(X):
        if X.ndim == 4:
            batch_size, num_bands, num_rows, num_cols = X.shape
            X = torch.permute(X, (0, 3, 2, 1))
            return X.reshape(-1, num_cols, num_rows * num_bands)
        else:
            num_bands, num_rows, num_cols = X.shape
            X = torch.permute(X, (2, 1, 0))
            return X.reshape(num_cols, num_rows * num_bands)
    else:
        if X.ndim == 4:
            batch_size, num_rows, num_cols, num_bands = X.shape
            X = np.transpose(X, (0, 2, 1, 3))
            return X.reshape(-1, num_cols, num_rows * num_bands)
        else:
            num_rows, num_cols, num_bands = X.shape
            X = np.transpose(X, (1, 0, 2))
            return X.reshape(num_cols, num_rows * num_bands)
        
            
def Fold2(X1, num_rows, num_bands):
    if X1.ndim == 3:
        num_cols = X1.shape[1]
    else:
        num_cols = X1.shape[0]

    if torch.is_tensor(X1):
        if X1.ndim == 3:
            X = X1.reshape(-1, num_cols, num_rows, num_bands)
            X = torch.permute(X, (0, 3, 2, 1))
            return X
        else:
            X = X1.reshape(num_cols, num_rows, num_bands)
            X = torch.permute(X, (2, 1, 0))
            return X
    else:
        if X1.ndim == 3:
            X = X1.reshape(-1, num_cols, num_rows, num_bands)
            X = np.transpose(X, (0, 2, 1, 3))
            return X
        else:
            X = X1.reshape(num_cols, num_rows, num_bands)
            X = torch.permute(X, (1, 0, 2))
            return X
        

def Unfold3(X):
    if torch.is_tensor(X):
        if X.ndim == 4:
            batch_size, num_bands, num_rows, num_cols = X.shape
            X3 = X.reshape(batch_size, num_bands, num_rows * num_cols)
#             X3 = torch.clone(X3, memory_format = torch.contiguous_format)
            return X3
        else:
            num_bands, num_rows, num_cols = X.shape
            X3 = X.reshape(num_bands, num_rows * num_cols)
#             X3 = torch.clone(X3, memory_format = torch.contiguous_format)
            return X3
    else:
        if X.ndim == 4:
            batch_size, num_rows, num_cols, num_bands = X.shape
            X = np.transpose(X, (0, 3, 1, 2))
            X3 = X.reshape(batch_size, num_bands, num_rows * num_cols)
#             X3 = X3.copy()
            return X3
        else:
            num_rows, num_cols, num_bands = X.shape
            X = np.transpose(X, (2, 0, 1))
            X3 = X.reshape(num_bands, num_rows * num_cols)
#             X3 = X3.copy()
            return X3

        
def Fold3(X3, num_rows, num_cols):
    if X3.ndim == 3:
        num_bands = X3.shape[1]
    else:
        num_bands = X3.shape[0]
        
    if torch.is_tensor(X3):
        if X3.ndim == 3:
            X = X3.reshape(-1, num_bands, num_rows, num_cols)
#             X = torch.clone(X, memory_format = torch.contiguous_format)
            return X
        else:
            X = X3.reshape(num_bands, num_rows, num_cols)
#             X = torch.clone(X, memory_format = torch.contiguous_format)
            return X
    else:
        if X3.ndim == 3:
            X = X3.reshape(-1, num_bands, num_rows, num_cols)
            X = np.transpose(X, (0, 2, 3, 1))
#             X = X.copy()
            return X
        else:
            X = X3.reshape(num_bands, num_rows, num_cols)
            X = np.transpose(X, (1, 2, 0))
#             X = X.copy()
            return X

        
# def BicubicInterpolation(L, spatial_ratio):
#     if torch.is_tensor(L):
#         device = L.device
#         L = ToNumpy(L)
#         tensor = True
#     else:
#         tensor = False
#     
#     if L.ndim == 3:
#         L = np.expand_dims(L, 0)
#         squeeze = True
#     else:
#         squeeze = False
# 
#     num_patches = L.shape[0]
#     num_rows = L.shape[1] * spatial_ratio
#     num_cols = L.shape[2] * spatial_ratio
#     num_bands = L.shape[3]
#     
#     X0 = np.zeros((num_patches, num_rows, num_cols, num_bands))
# 
#     for patch in range(num_patches):
#         for band in range(num_bands):
#             X0[patch, :, :, band] = \
#                 np.array(Image.fromarray(L[patch, :, :, band].squeeze()).resize(size=(num_cols, num_rows), resample=Image.BICUBIC))
# 
#     if tensor:
#         X0 = ToTensor(X0).to(device)
#     if squeeze:
#         return X0.squeeze()
#     else:
#         return X0

    
def BicubicInterpolation(L, spatial_ratio):
    return_numpy = False
    if not torch.is_tensor(L):
        L = ToTensor(L)
        return_numpy = True

    return_squeeze = False
    if (len(L.shape)) == 3:
        L = torch.unsqueeze(L, 0)
        return_squeeze = True
        
    X0 = torch.nn.functional.interpolate(L,
                                         scale_factor = spatial_ratio,
                                         mode = "bicubic")
                                             
    if return_squeeze:
        X0 = torch.squeeze(X0, 0)
    if return_numpy:
        return ToNumpy(X0)
    else:
        return X0

# 
#     else:
#         if L.ndim == 3:
#             L = np.expand_dims(L, 0)
#             return_squeeze = True
#         else:
#             return_squeeze = False
# 
#         num_patches = L.shape[0]
#         num_rows = L.shape[1] * spatial_ratio
#         num_cols = L.shape[2] * spatial_ratio
#         num_bands = L.shape[3]
#     
#         X0 = np.zeros((num_patches, num_rows, num_cols, num_bands))
# 
#         for patch in range(num_patches):
#             for band in range(num_bands):
#                 X0[patch, :, :, band] = \
#                     np.array(Image.fromarray(L[patch, :, :, band].squeeze()).resize(size=(num_cols, num_rows), resample=Image.BICUBIC))
#             
#         if return_squeeze:
#             return X0.squeeze()
#         else:
#             return X0


def GetPCATransform(X, num_pcs = 'all'):
    if torch.is_tensor(X):
        dtype = X.dtype
        device = X.device
        X = ToNumpy(X)
        tensor = True
    else:
        tensor = False

    L, U = np.linalg.eigh(Unfold3(X) @ Unfold3(X).T)
    if num_pcs != 'all':
        U = U[..., :num_pcs]

    if tensor:
        U = torch.tensor(U, dtype = dtype, device = device)
    return U


def Tile(X, padding, tile_shape, overlap):
    X = np.pad(X, (padding, padding, (0, 0)), mode='reflect')
    X_tiler = tiler.Tiler(data_shape = X.shape,
                          tile_shape = tile_shape,
                          overlap = overlap,
                          channel_dimension = 2)
    X_tiled = np.zeros((len(X_tiler),) + tile_shape, dtype = np.float32)
    X_tiler = Tiler(X, padding, tile_shape, overlap)
    X_tiled = np.zeros((len(X_tiler),) + tile_shape, dtype = np.float32)
    for tile_id, tile in X_tiler.iterate(X):
        X_tiled[tile_id, :, :, :] = tile

    return X_tiled, X_tiler


def MSE(X, X_hat):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)
    return skim.mean_squared_error(X, X_hat)


def RMSE(X, X_hat):
    return np.sqrt(MSE(X, X_hat))


def PSNR(X, X_hat, max_val = 255.):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)
    return skim.peak_signal_noise_ratio(X, X_hat, data_range = max_val)


def SNR(X, X_hat):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)
    mse = MSE(X, X_hat)
    if mse > 0:
        var = np.var(X)
        snr = 10 * np.log10(var / mse)
    else:
        snr = np.nan
    return snr

    
# def SAM(X, X_hat):
#     X = ToNumpy(X)
#     X_hat = ToNumpy(X_hat)
#     num_rows, num_cols, num_bands = X.shape
#     s = np.zeros((num_rows, num_cols))
#     for row in range(num_rows):
#         for col in range(num_cols):
#             x = X[row, col, :].squeeze()
#             x_hat = X_hat[row, col, :].squeeze()
#             x_norm = npla.norm(x)
#             x_hat_norm = npla.norm(x_hat)
#             if (x_norm < 1e-7) or (x_hat_norm < 1e-7):
#                 s[row, col] = np.nan
#             else:
#                 s[row, col] = np.real(np.arccos(np.dot(x, x_hat) /
#                                                 x_norm /
#                                                 x_hat_norm)) / \
#                                                 np.pi * 180
#     return np.mean(s)


def SAM(X, X_hat):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)
    X_norm = npla.norm (X, axis = -1)
    X_hat_norm = npla.norm (X_hat, axis = -1)
    I = (X_norm >= 1e-7) & (X_hat_norm >= 1e-7)
    s = np.zeros_like(X_norm)
    dp = np.einsum("...i,...i", X, X_hat)
    s[I] = dp[I] / X_norm[I] / X_hat_norm[I]
    s = np.clip(s, -1, 1)
    s[I] = np.real(np.arccos(s[I])) / np.pi * 180
    return np.mean(s[I])


def SSIM(X, X_hat, max_val = 255.):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)
    X = np.clip(X, 0.0, 1.0)
    X_hat = np.clip(X_hat, 0.0, 1.0)
    ssim = skim.structural_similarity(X, X_hat,
                                      data_range = max_val, channel_axis = -1)
    return ssim


# def TV(X):
#     if len(X.shape) == 3:
#         X = tf.expand_dims(X, axis = 0)
#     if len(X_hat.shape) == 3:
#         X_hat = tf.expand_dims(X_hat, axis = 0)
#     return tf.image.total_variation(X)
# 
# def nuclear_norm(X, U):
#     if (len(X.shape) == 3):
#         X = tf.expand_dims(X, axis = 0)
#     if (len(U.shape) == 2):
#         U = tf.expand_dims(U, axis = 0)
#     X3 = Unfold3(X)
#     Z = Unfold3(tfla.matmul(X3, U, adjoint_a = True))
#     S2 = tfla.matmul(Z, Z, adjoint_b = True)
#     
#     tf.debugging.check_numerics(S2, 'problematic S2')
#     return tfla.trace(tf.math.sqrt(S2))


def GetErrorPlot(X, X_hat, return_colormap = True, location = "bottom"):
    X = ToNumpy(X)
    X_hat = ToNumpy(X_hat)

    diff = np.sqrt(np.mean((X - X_hat) ** 2, axis = -1))
#     j = plt.imshow(diff, vmin = 0., vmax = 0.2, cmap='viridis')
    j = plt.imshow(diff, vmin = 0., vmax = 0.2, cmap='nipy_spectral')
    plt.gca().set_axis_off()
    plot = plt.gcf()
    if not return_colormap:
        return plot
    _, ax = plt.subplots()
    plt.colorbar(j, ax = ax, location = location)
    ax.remove()
    colorbar = plt.gcf()
    return plot, colorbar


def SaveErrorPlot(X, X_hat, errfile, colorbarfile = None, location = "bottom"):
    if colorbarfile is None:
        plot = GetErrorPlot(X, X_hat, return_colormap = False)
        plot.savefig(errfile, bbox_inches = 'tight', pad_inches = 0)
    else:
        plot, colorbar = GetErrorPlot(X, X_hat, location = location)
        plot.savefig(errfile, bbox_inches = 'tight', pad_inches = 0)
        colorbar.savefig(colorbarfile, bbox_inches = 'tight', pad_inches = 0)


def ProcessSensor(sensor = None):
    remove_absorption_bands = None
    if sensor == 'aviris':
        S = aviris.GetSpectralProjection().astype(np.float32)
        remove_absorption_bands = aviris.RemoveAbsorptionBands
        get_pseudocolor_image = aviris.GetPseudocolorImage
    elif sensor == 'aviris220':
        S = aviris220.GetSpectralProjection().astype(np.float32)
        remove_absorption_bands = aviris220.RemoveAbsorptionBands
        get_pseudocolor_image = aviris220.GetPseudocolorImage
    elif sensor == 'rosis':
        S = rosis.GetSpectralProjection().astype(np.float32)
        remove_absorption_bands = rosis.RemoveAbsorptionBands
        get_pseudocolor_image = rosis.GetPseudocolorImage
    elif sensor == 'vnir':
        S = vnir.GetSpectralProjection().astype(np.float32)
        remove_absorption_bands = vnir.RemoveAbsorptionBands
        get_pseudocolor_image = vnir.GetPseudocolorImage
    elif sensor == 'hydice':
        S = hydice.GetSpectralProjection().astype(np.float32)
        remove_absorption_bands = hydice.RemoveAbsorptionBands
        get_pseudocolor_image = hydice.GetPseudocolorImage
    else:
        print('Invalid sensor type (%s)' % sensor, flush = True)
        return None
    return S, remove_absorption_bands, get_pseudocolor_image
    



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import subprocess
import glob
import numpy as np
import hdf5storage as hdf
import imageio
from Utils import rosis

def MakeData():
    dirname = os.path.dirname(os.path.realpath(__file__))

    original_dir = os.path.join(dirname, "Original")
    os.makedirs(original_dir, exist_ok = True)

    original_file = os.path.join(original_dir, "pavia.mat")
    original_image = os.path.join(original_dir, "pavia.original.png")

    if os.path.exists(original_file):
        print("  Nothing to do - %s exists already" % original_file)
        return
    
    fileurl = "https://www.ehu.eus/ccwintco/uploads/e/ee/PaviaU.mat"
    print("Downloading %s..." % fileurl)
    if subprocess.run(['wget', '-q', '-O', original_file, fileurl]):
        print("Error downloading %s" % fileurl)
        return

    training_dir = os.path.join(dirname, "Training")
    os.makedirs(training_dir, exist_ok = True)
    training_file = os.path.join(training_dir, "pavia.training.mat")
    training_image = os.path.join(training_dir, "pavia.training.png")

    testing_dir = os.path.join(dirname, "Testing")
    os.makedirs(testing_dir, exist_ok = True)
    testing_file = os.path.join(testing_dir, "pavia.testing.mat")
    testing_image = os.path.join(testing_dir, "pavia.testing.png")
    
    X_original = hdf.loadmat(original_file)['X']
    outimage = rosis.GetPseudocolorImage(X_original)
    imageio.imwrite(original_image, outimage)
    
    # Training: lower-left 256x256
    X = X_original[-256:, :256, :]
    if os.path.isfile(training_file):
        os.remove(training_file)
    hdf.savemat(training_file, {'X': X})
    outimage = rosis.GetPseudocolorImage(X)
    imageio.imwrite(training_image, outimage)
    
    # Testing: upper-left 256x256
    X = X_original[:256, :256, :]
    if os.path.isfile(testing_file):
        os.remove(testing_file)
    hdf.savemat(testing_file, {'X': X})
    outimage = rosis.GetPseudocolorImage(X)
    imageio.imwrite(testing_image, outimage)



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import tempfile
import zipfile
import shutil
import subprocess
import glob
import numpy as np
import hdf5storage as hdf
import imageio
from Utils import vnir

def MakeData():
    num_rows_extracted = 256
    num_cols_extracted = 256
    origin_row = 300
    origin_col = 1700

    dirname = os.path.dirname(os.path.realpath(__file__))

    original_dir = os.path.join(dirname, "Original")
    os.makedirs(original_dir, exist_ok = True)

    original_file = os.path.join(original_dir, "chikusei.mat")
    original_image = os.path.join(original_dir, "chikusei.original.png")

    fileurl = "http://park.itc.u-tokyo.ac.jp/sal/hyperdata/Hyperspec_Chikusei_MATLAB.zip"
    zipdir = "Chikusei_MATLAB"
    matfile = "HyperspecVNIR_Chikusei_20140729.mat"

    if os.path.exists(original_file):
        print("  Nothing to do - %s exists already" % original_file)
        return

    with tempfile.TemporaryDirectory() as tmp_dir:
        zfn = os.path.join(tmp_dir, os.path.basename(fileurl))
        print("Downloading %s as %s..." % (fileurl, zfn))
        if subprocess.run(['wget', '-q', '-P', tmp_dir, fileurl]):
            print("Error downloading %s" % fileurl)
            return
        zf = zipfile.ZipFile(zfn)
        print("Extracting %s..." % matfile)
        zf.extract("%s/%s" % (zipdir, matfile), path = tmp_dir)
        X_original = hdf.loadmat(os.path.join(tmp_dir, zipdir, matfile))["chikusei"].astype(np.float32)
        print("Saving %s..." % original_file)
        hdf.savemat(original_file, {'X': X_original.astype(np.uint16)})

    training_dir = os.path.join(dirname, "Training")
    os.makedirs(training_dir, exist_ok = True)
    training_file = os.path.join(training_dir, "chikusei.training.mat")
    training_image = os.path.join(training_dir, "chikusei.training.png")

    testing_dir = os.path.join(dirname, "Testing")
    os.makedirs(testing_dir, exist_ok = True)
    testing_file = os.path.join(testing_dir, "chikusei.testing.mat")
    testing_image = os.path.join(testing_dir, "chikusei.testing.png")
    
    outimage = vnir.GetPseudocolorImage(X_original)
    imageio.imwrite(original_image, outimage)
    
    # Training: upper-right 256x256
    X = X_original[origin_row:(origin_row + num_rows_extracted),
                               origin_col:(origin_col + num_cols_extracted), :]
    if os.path.isfile(training_file):
        os.remove(training_file)
    hdf.savemat(training_file, {'X': X})
    outimage = vnir.GetPseudocolorImage(X)
    imageio.imwrite(training_image, outimage)
    
    # Testing: lower-right 256x256
    X = X_original[-origin_row:(-origin_row + num_rows_extracted),
                   origin_col:(origin_col + num_cols_extracted), :]
    if os.path.isfile(testing_file):
        os.remove(testing_file)
    hdf.savemat(testing_file, {'X': X})
    outimage = vnir.GetPseudocolorImage(X)
    imageio.imwrite(testing_image, outimage)



# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import numpy as np
import torch
from numpy import linalg as npla
import warnings
warnings.filterwarnings('ignore')
from oct2py import octave


def svt(Z, tau):
    m = Z.shape[-2]
    n = Z.shape[-1]
    if m < n:
        ZZH = Z @ Z.adjoint()
        S2, U = torch.linalg.eigh(ZZH)
        I = S2 > torch.square(tau)
        S = torch.where(I, torch.sqrt(S2), 0.)
        S[I] = (S[I] - tau) / S[I]
        S = torch.diag_embed(S)
        return U @ S @ U.adjoint() @ Z
    else:
        ZZH = Z.adjoint() @ Z
        S2, U = torch.linalg.eigh(ZZH)
        I = S2 > torch.square(tau)
        S = torch.where(I, torch.sqrt(S2), 0.)
        S[I] = (S[I] - tau) / S[I]
        S = torch.diag_embed(S)
        return Z @ U @ S @ U.adjoint()
 

def svt_fixedbasis(Z, tau, U):
    m = Z.shape[-2]
    n = Z.shape[-1]
    if m < n:
        S = torch.linalg.vector_norm(U.adjoint() @ Z, dim = -1)
        I = S > tau
        S = torch.where(I, S, 0.)
        S[I] = (S[I] - tau) / S[I]
        S = torch.diag_embed(S)
        return U @ S @ U.adjoint() @ Z
    else:
        S = torch.linalg.vector_norm(Z @ U, dim = -2)
        I = S > tau
        S = torch.where(I, S, 0.)
        S[I] = (S[I] - tau) / S[I]
        S = torch.diag_embed(S)
        return Z @ U @ S @ U.adjoint()
        



#!/usr/bin/env python3

# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import os
import sys
import torch
import argparse
from CCDD import ccdd_reconstruct as ccdd


##########################################################


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--sensor',
                        default = 'aviris',
                        help='(string) sensor type (default = "aviris")')
    parser.add_argument('-sr', '--spatial_ratio',
                        type = int,
                        default = 8,
                        help = '(int) spatial subsampling ratio (default = 8)')
    parser.add_argument('-bb', '--band_block',
                        type = int,
                        default = 3,
                        help = '(int) number of bands to block together (default = 3)')
    parser.add_argument('-T', '--num_train_steps',
                        type = int,
                        default = 1000,
                        help = '(int) number of diffusion time steps (default = 1000)')
    parser.add_argument('-bmx', '--beta_max',
                        type = float,
                        default = 0.01,
                        help = '(float) maximum beta value (default = 0.01)')
    parser.add_argument('-bmn', '--beta_min',
                        type = float,
                        default = 0.0001,
                        help = '(float) minimum beta value (default = 0.0001)')
    parser.add_argument('-ns', '--num_inference_steps',
                        type = int,
                        default = 10,
                        help = '(int) number of inference time steps (default = 10)')
    parser.add_argument('-li', '--num_landweber_iterations',
                        type = int,
                        default = 2,
                        help = '(int) number of Landweber iterations (default = 5)')
    parser.add_argument('--ddim', action = 'store_true',
                        default = False,
                        help = 'use DDIM scheduler (default = DDPM)')
    parser.add_argument('--unconstrained', action = 'store_true',
                        default = False,
                        help = 'no enforcement of S or P constraints (implies Landweber iterations = 1)')
    parser.add_argument('--unconditioned', action = 'store_true',
                        default = False,
                        help = 'no conditioning')
    parser.add_argument('--transpose_adjoint', action = 'store_true',
                        default = False,
                        help = 'use transpose as spatial adjoint operator (default = bicubic interpolation)')
    parser.add_argument('--tau',
                        type = float,
                        default = None,
                        help = '(float) spectral regularizer constant (default = 3/2/sigma_1^2, where sigma_1 is the largest singular value of the S matrix)')
    parser.add_argument('--eta',
                        type = float,
                        default = 1.0,
                        help = '(float) spatial regularizer constant (default = 1.0')
    parser.add_argument('--flops', action = 'store_true',
                        default = False,
                        help = 'report number of FLOPS')
    parser.add_argument('testing_dir',
                        help = '(string) directory of testing images (MAT format)')
    parser.add_argument('model_dir',
                        help='(string) directory of trained model')
    parser.add_argument('results_dir',
                        help='(string) directory of output results')
    args = parser.parse_args()
    return args


##########################################################


def main():

    args = parse_arguments()

    device_name = "cuda" if torch.cuda.is_available() else "cpu"
    device = torch.device(device_name)

    ccdd.CCDDReconstruct(
        testing_dir = args.testing_dir,
        model_dir = args.model_dir,
        results_dir = args.results_dir,
        sensor = args.sensor,
        spatial_ratio = args.spatial_ratio,
        band_block = args.band_block,
        num_train_steps = args.num_train_steps,
        beta_max = args.beta_max,
        beta_min = args.beta_min,
        num_inference_steps = args.num_inference_steps,
        num_landweber_iterations = args.num_landweber_iterations,
        tau = args.tau,
        eta = args.eta,
        ddim = args.ddim,
        unconstrained = args.unconstrained,
        unconditioned = args.unconditioned,
        transpose_adjoint = args.transpose_adjoint,
        flops = args.flops,
        device = device,
        quiet = False)

    
##########################################################


if __name__ == "__main__":
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, dirname)
    main()



#!/usr/bin/env python3

# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 

import os
import sys
import torch
import argparse
from CCDD import ccdd_train as ccdd


##########################################################


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--sensor', default = 'aviris',
                        help='(string) sensor type (default = "aviris")')
    parser.add_argument('-T', '--num_train_steps', type=int, default=1000,
                        help = '(int) number of diffusion time steps (default = 1000)')
    parser.add_argument('-bmx', '--beta_max', type=float, default=0.01,
                        help = '(float) maximum beta value (default = 0.01)')
    parser.add_argument('-bmn', '--beta_min', type=float, default=0.0001,
                        help = '(float) minimum beta value (default = 0.0001)')
    parser.add_argument('-lr', '--learning_rate', type=float, default=0.0001,
                        help = '(float) learning rate of ADAM optimizer (default = 0.0001')
    parser.add_argument('-bs', '--batch_size', type=int, default=64,
                        help = '(int) training batch size (default = 64)')
    parser.add_argument('-ps', '--patch_size', type=int, default=32,
                        help = '(int) patch size (default = 32)')
    parser.add_argument('-np', '--num_patches', type=int, default=8192,
                        help = '(int) number of patches per image (default = 8192)')
    parser.add_argument('-bb', '--band_block', type=int, default=3,
                        help = '(int) number of bands to block together (default = 3)')
    parser.add_argument('-uc', '--unet_channels', type=int, default=128,
                        help = '(int) UNet input channels (default = 128)')
    parser.add_argument('-ul', '--unet_layers', type=int, default=3,
                        help = '(int) num UNet layers (default = 3)')
    parser.add_argument('-e', '--epochs', type=int, default=300,
                        help = '(int) number of training epochs (default = 300)')
    parser.add_argument('--unconditioned', action = 'store_true',
                        default = False,
                        help = 'no conditioning')
    parser.add_argument('-sr', '--spatial_ratio', type=int, default=8,
                        help = '(int) spatial subsampling ratio (default = 8)')
    parser.add_argument('-r', '--resume', type=int, default=0,
                        help = '(int) resume training at given epoch')
    parser.add_argument('-se', '--save_epochs', type=int, default=10,
                        help = '(int) epoch interval for saving model')
    parser.add_argument('-hs', '--history_file', type=str, default=None,
                        help = '(string) filename of output training history (default is do not save history')
    parser.add_argument('training_dir',
                        help = '(string) directory of input training files')
    parser.add_argument('model_dir',
                        help='(string) directory of output model')
    args = parser.parse_args()
    return args


##########################################################


def main():

    args = parse_arguments()

    device_name = "cuda" if torch.cuda.is_available() else "cpu"
    device = torch.device(device_name)

    model_unet, optimizer, losses = \
        ccdd.CCDDTrain(training_dir = args.training_dir,
                       model_dir = args.model_dir,
                       sensor = args.sensor,
                       spatial_ratio = args.spatial_ratio,
                       band_block = args.band_block,
                       num_train_steps = args.num_train_steps,
                       beta_max = args.beta_max,
                       beta_min = args.beta_min,
                       learning_rate = args.learning_rate,
                       batch_size = args.batch_size,
                       patch_size = args.patch_size,
                       num_patches = args.num_patches,
                       unet_channels = args.unet_channels,
                       unet_layers = args.unet_layers,
                       start_epoch = args.resume,
                       end_epoch = args.epochs,
                       save_epochs = args.save_epochs,
                       history_file = args.history_file,
                       unconditioned = args.unconditioned,
                       device = device,
                       quiet = False)

    
##########################################################


if __name__ == "__main__":
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, dirname)
    main()



#!/usr/bin/env python3

# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import sys
from Data.Pavia import make_data as pavia
from Data.Chikusei import make_data as chikusei
from Data.LosAngeles import make_data as losangeles


def MakeData():
    print("Processing Pavia University")
    pavia.MakeData()
    print("Processing Chikusei")
    chikusei.MakeData()
    print("Processing Los Angeles")
    losangeles.MakeData()


if __name__ == "__main__":
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, dirname)
    MakeData()
